use crate::prelude::*;

pub(crate) struct Args {
    pub(crate) name: String,
    pub(crate) port: u16,
}

impl Default for Args {
    fn default() -> Self {
        Self {
            name: "anon".to_string(),
            port: 9000,
        }
    }
}

pub(crate) struct App {
    client: Client,
    stream: Framed<TcpStream, LengthDelimitedCodec>,
}

impl App {
    pub(crate) async fn new(args: Args) -> Result<Self> {
        let stream = TcpStream::connect(("127.0.0.1", args.port)).await?;
        let stream = Framed::new(stream, LengthDelimitedCodec::new());
        Ok(Self {
            client: Client::new(args.name)?,
            stream,
        })
    }

    pub(crate) fn handle_stdin(&mut self, line: String) -> Result<()> {
        self.client.handle_stdin(line)
    }

    pub(crate) fn poll_run(&mut self, cx: &mut Context<'_>) -> Poll<Result<()>> {
        match self.step(cx) {
            Ok(()) => Poll::Pending,
            Err(err) => Poll::Ready(Err(err)),
        }
    }

    fn step(&mut self, cx: &mut Context<'_>) -> Result<()> {
        let mut stream = pin!(&mut self.stream);
        if let Poll::Ready(result) =
            <_ as futures_sink::Sink<Bytes>>::poll_ready(stream.as_mut(), cx)
        {
            result?;
            if let Some(frame) = self.client.poll_send() {
                stream
                    .as_mut()
                    .start_send(frame)
                    .context("stream.start_send")?;
                tracing::debug!("Started send");
            }
            match <_ as futures_sink::Sink<Bytes>>::poll_flush(stream.as_mut(), cx) {
                Poll::Pending => {}
                Poll::Ready(result) => {
                    result.context("poll_flush")?;
                }
            }
        }

        if let Poll::Ready(frame_opt) = stream.as_mut().poll_next(cx) {
            let frame = frame_opt.context("Server closed cxn")?;
            cx.waker().wake_by_ref();
            let frame = frame.context("network framing decode")?;
            let output = self
                .client
                .handle_frame(frame.into())
                .context("client.handle_frame")?;
            println!("{output}");
        }

        Ok(())
    }
}

#[derive(Default)]
struct Client {
    outbox: VecDeque<Bytes>,
    sequence: u64,
}

impl Client {
    fn new(name: String) -> Result<Self> {
        let mut this = Self {
            outbox: Default::default(),
            sequence: 0,
        };

        this.handle_set_name(name)?;
        Ok(this)
    }

    fn enqueue(&mut self, msg: &ToServer) -> Result<()> {
        if !self.outbox.is_empty() {
            bail!("Dropped message, outbox is full");
        }
        let frame = rmp_serde::to_vec(&msg)?;
        self.outbox.push_back(frame.into());
        Ok(())
    }

    fn handle_frame(&self, frame: Bytes) -> Result<String> {
        let ToClient { id, name, event } = rmp_serde::from_slice(&frame)?;
        Ok(match event {
            ToClientEvent::ChatLine { line } => format!("{name}: {line}"),
            ToClientEvent::Connected => format!("New client ({id}) connected"),
            ToClientEvent::Disconnected => format!("{name} ({id}) disconnected"),
            ToClientEvent::NameChanged { old_name } => {
                format!("\"{old_name}\" ({id}) changed their name to \"{name}\"")
            }
        })
    }

    fn handle_set_name(&mut self, name: String) -> Result<()> {
        self.enqueue(&ToServer::SetName { name })
    }

    pub(crate) fn handle_stdin(&mut self, line: String) -> Result<()> {
        let msg = ToServer::ChatLine {
            line,
            sequence: self.sequence,
        };
        self.sequence += 1;
        self.enqueue(&msg)
    }

    fn poll_send(&mut self) -> Option<Bytes> {
        self.outbox.pop_front()
    }
}
